# -*- dbm: python ; coding: utf-8 -*-
import re
from datetime import datetime, date, time, timedelta
from decimal import Decimal
from typing import Callable

import jinja2
import pandas as pd
from sqlalchemy import create_engine, MetaData, inspect, Integer, BigInteger, SmallInteger, Float, Numeric
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.orm import sessionmaker, Query
from sqlalchemy.sql.elements import quoted_name
from sqlalchemy.types import TypeDecorator

from yut import CaseInsensitiveDict, Obj, is_collection, o2d, is_execute_able, get_attr, set_attr, call_exp
from . import Op, QueryR


def extract_qc(params: dict | object | Obj, name: str) -> None | Obj:
    """
    从字典或对象中提取指定的查询条件值
    :param params: 包含多个查询条件的字典或对象，字典的格式为 name: (op,value)
    :param name: 条件名称
    :return: 如果没有对应名称的条件，返回None，否则返回Obj对象，包含attr,op,value三个属性.
    """
    v = get_attr(params, name)
    if v:
        if is_collection(v):
            op, value = v  # Op , value
        else:
            op, value = Op.EQU, v
        return Obj(attr=name, op=Op(op), value=value)
    return None


def qc_value(params: dict | object | Obj, name: str, none_as=None, take=False) -> None | Obj:
    value = get_attr(extract_qc(params, name), 'value')
    if take:
        set_attr(params, name, value)
    return value if value is not None else none_as


INTEGER_TYPES = (Integer, BigInteger, SmallInteger)


class PandasTypeMapper:
    """
    SQLAlchemy 类型到 pandas 类型的映射器
    """

    BASE_MAPPING = {
        Integer: pd.Int64Dtype(),
        BigInteger: pd.Int64Dtype(),
        SmallInteger: pd.Int64Dtype(),
        Float: float,
        Numeric: float,  # 或 'object' 保留 Decimal
    }

    # 特殊类型处理
    @staticmethod
    def map_type(sqlalchemy_type):
        """
        映射 SQLAlchemy 类型到 pandas 类型
        """
        """映射 SQLAlchemy 类型到 pandas 类型（安全处理字符串）"""
        # 检查直接匹配
        for base_type, pandas_type in PandasTypeMapper.BASE_MAPPING.items():
            if isinstance(sqlalchemy_type, base_type):
                return pandas_type

        # 处理 TypeDecorator
        if isinstance(sqlalchemy_type, TypeDecorator):
            return PandasTypeMapper.map_type(sqlalchemy_type.impl)

        # 处理具有 python_type 的类型
        if hasattr(sqlalchemy_type, 'python_type'):
            py_type = sqlalchemy_type.python_type
            if py_type == int:
                return pd.Int64Dtype()
            elif py_type == float:
                return float
            elif py_type == bool:
                return pd.BooleanDtype()
            elif py_type == str:
                return 'safe_string'  # 特殊标记
            elif py_type == datetime:
                return 'datetime64[ns]'
            elif py_type == date:
                return 'datetime64[ns]'
            elif py_type == time:
                return 'timedelta64[ns]'
            elif py_type == timedelta:
                return 'timedelta64[ns]'
            elif py_type == Decimal:
                return float

        return 'object'

    @staticmethod
    def get_column_types(query):
        """
        从 SQLAlchemy 查询中获取列类型映射
        """
        col_types = {}
        try:
            # 现代 SQLAlchemy (1.4+) 方式
            if hasattr(query, 'selected_columns'):
                for col in query.selected_columns:
                    col_types[col.key] = PandasTypeMapper.map_type(col.type)
            # 旧版本 SQLAlchemy 方式
            else:
                for col_desc in query.column_descriptions:
                    col_name = col_desc['name']
                    col_type = col_desc['type']
                    col_types[col_name] = PandasTypeMapper.map_type(col_type)
        except Exception as e:
            print(f"Warning: Failed to get column types - {str(e)}")
        return col_types

    @staticmethod
    def safe_string_conversion(series):
        """
        安全转换字符串类型，处理空值
        """
        # 处理空值
        if series.isnull().any():
            # 创建新的 StringArray 处理缺失值
            return pd.array(series.astype(str).replace('nan', pd.NA), dtype=pd.StringDtype())

        # 无空值时直接转换
        return series.astype(str)

    @staticmethod
    def convert_to_pandas_type(series, target_type):
        """
        安全转换 pandas 系列到目标类型
        """
        if target_type == 'safe_string':
            return PandasTypeMapper.safe_string_conversion(series)

        # 特殊处理日期时间类型
        if target_type == 'datetime64[ns]':
            return pd.to_datetime(series, errors='coerce')

        # 特殊处理时间差类型
        if target_type == 'timedelta64[ns]':
            if series.dtype == object:
                # 转换 time 对象为 timedelta
                return series.apply(
                    lambda x: pd.Timedelta(
                        hours=x.hour,
                        minutes=x.minute,
                        seconds=x.second
                    ) if isinstance(x, time) else pd.NaT
                )
            return series.astype('timedelta64[ns]', errors='ignore')

        # 标准转换
        try:
            return series.astype(target_type)
        except (TypeError, ValueError):
            return series.astype('object')

    @staticmethod
    def dataframe_as_types(df, col_types):
        # 应用精确类型转换
        for col_name in df.columns:
            if col_name in col_types:
                target_type = col_types[col_name]
                df[col_name] = PandasTypeMapper.convert_to_pandas_type(df[col_name], target_type)
        return df


def field_types(query):
    """
    从 SQLAlchemy 查询中获取列类型映射
    """
    try:
        # 现代 SQLAlchemy (1.4+) 方式
        if hasattr(query, 'selected_columns'):
            return {col.key: col.type for col in query.selected_columns}
        # 旧版本 SQLAlchemy 方式
        else:
            return {col_desc['name']: col_desc['type'] for col_desc in query.column_descriptions}
    except Exception as e:
        print(f"Warning: Failed to get column types - {str(e)}")
        return {}


class Selection:
    def __init__(self,
                 db,
                 title=None,
                 comment=None,
                 tables=None,
                 columns=None,
                 joins=None,
                 group_bys=None,
                 order_bys=None,
                 make_q=None,
                 convert_rec=None,
                 post_process=None,
                 ext_comments: dict = None,
                 unit_spec: dict = None,
                 ):
        self.db = db
        self.title = title
        self.comment = comment
        self.session = None
        self.tables = tables
        self.columns = columns
        self.joins = joins
        self.group_bys = group_bys
        self.order_bys = order_bys
        self.make_q = make_q
        self.convert_record = convert_rec
        self.post_process = post_process
        self.ext_comments = ext_comments if ext_comments else {}  # 附加的备注，如用在sum、count等聚合函数生成的字段
        self.unit_spec = unit_spec
        self._cls_dict = {}
        self.query_args = []
        self.query_kwargs = {}
        self._conditions = CaseInsensitiveDict()  # 专门使用的查询条件，形如 "key": (Op,value)
        self._result = None
        self.setup()

    def setup(self):
        # 创建model_class字典
        self._cls_dict = {chr(ord('a') + i): self.db.model_class(name) for i, name in enumerate(self.tables)}

    def set_make_q(self, make_q):
        self.make_q = make_q

    def set_convert_rec(self, rc):
        self.convert_record = rc

    def set_post_process(self, pp):
        self.post_process = pp

    def _column_list(self):
        return [self.f(col) for col in self.columns]

    def add_ext_comments(self, **kwargs):
        self.ext_comments.update(**kwargs)

    def add_unit_spec(self, col, **kwargs):
        self.unit_specs[col] = kwargs

    def column_spec(self) -> list:
        def col_comment(col):
            if self.ext_comments.get(col.key):  # 优先使用ext_comments,以便局部修改默认值
                return self.ext_comments[col.key]
            attr = 'comment'
            if hasattr(col, attr):
                return getattr(col, attr)
            return None

        ret = [{'key': c.key,
                'name': c.name,
                'title': col_comment(c),# 反射出的字段描述通常放在comment中，赋值给title属性
                'comment': col_comment(c),
                'dtype': f"{c.type}" if c.type is not None else None,
                'table': f"{c.table}" if c.table is not None else None,
                } for c in self._column_list()]

        # 允许ext_comment中补充不在columns中的列附注，如后处理时对dataframe增加列，因此需要添加ext_comment中没有包含在ret中的元素
        for ec_key, ec_value in self.ext_comments.items():
            if ec_key not in [r['key'] for r in ret]:
                ret.append({'key': ec_key, 'name': ec_key, 'comment': ec_value, 'dtype': None, 'table': None})

        # 补充设置 unit spec
        for c in ret:
            if self.unit_spec:
                u_spec = self.unit_spec.get(c['key'])
                if u_spec:
                    c.update(u_spec)  # 将u_spec中的key-value平铺到 col_spec 中

        return ret

    def t_keys(self):
        return self._cls_dict.keys()

    def cls_attr(self, t_key, attr):
        for cls in [self.t(t_key)] if t_key else self._cls_dict.values():  # 给了表标识,明确使用，否则取第一个匹配的列名
            # 忽略大小写，逐个判断
            for key in cls.__dict__:
                if key.lower() == attr.lower().strip():
                    return getattr(cls, key)
        print(f'MetaDb.cls_attr(): attribute access failed , no such attr: {t_key} - {attr}')
        return None

    def t(self, key):
        r = self._cls_dict.get(key)
        if r:
            return r
        elif isinstance(key, int):  # 也允许使用下标索引
            return list(self._cls_dict.values())[key]
        elif key in self.tables:  # 直接使用表名也可以
            index = self.tables.index(key)
            return list(self._cls_dict.values())[index]
        return None

    def f(self, element):
        # 传入的是可调用对象，执行调用，返回结果
        if callable(element):  # 允许使用lambda动态返回属性对象(如使用func.sum().label()等)，也允许使用字符列名
            return element(self)
        if type(element) is quoted_name:
            element = str(element)
        if type(element) is not str:  # 传入的不是字符串，直接返回
            return element
        tokens = element.split('.')
        if len(tokens) == 1:  # 没有表前缀
            return self.cls_attr(t_key=None, attr=tokens[0].strip())
        else:
            return self.cls_attr(tokens[0].strip(), tokens[1].strip())

    def result(self):
        return self._result

    def __getattr__(self, item):
        cls = self._cls_dict.get(item)
        return cls if cls else object.__getattribute__(self, item)

    def push_conditions(self, o, clear=True):
        if clear:
            self._conditions.clear()
        if o is None:
            return
        if isinstance(o, dict):
            self._conditions.update(o)
        else:
            self._conditions.update(o.__dict__)

    def get_conditions(self) -> list[Obj]:
        ret = [self.get_condition(key) for key in self._conditions]
        return [o for o in ret if o is not None]

    def get_condition(self, attr: str):
        return extract_qc(self._conditions, attr)

    def add_join(self, q, *f_exps, outer=False):
        """
        添加表间连接
        :param q:
        :param f_exps:
        :param outer: 是否为 outer join，默认为 False，创建 inner join 。
        :return:
        """
        lgs = f_exps if len(f_exps) > 1 else [  # 只给出了一个参数，默认关联前两张表，字段名相同
            list(self.t_keys())[0] + f".{f_exps[0]}",
            list(self.t_keys())[1] + f".{f_exps[0]}"]
        lg = [self.f(ex) for ex in lgs]
        args = [f"(lg[{i * 2}] == lg[{i * 2 + 1}])" for i in range(int(len(lgs) / 2))]
        fn = q.outerjoin if outer else q.join
        return fn(lg[1].table, eval('&'.join(args)))

    def add_filter(self, query, column, value=None, op: Op = Op.EQU):
        if column is None:
            return query
        field = self.f(column)
        if not field:
            print(f'[Selection.add_filter]: 没有找到{column}对应的属性,也不是有效的表达式，忽略filter条件。')
            return query
        if value is None:
            return query
        operator = op if isinstance(op, Op) else Op(op)
        return operator.filter_q(query, field, value)

    def _create_query(self) -> Query:
        if self.session is None:
            self.session = self.db.Session()
        q = self.session.query()
        if self.columns:
            for col in self.columns:
                q = q.add_column(self.f(col))
        else:
            for key_table in self.tables:
                q = q.add_entity(self.t(key_table))
        if self.joins:
            for join in self.joins:
                if type(join[-1]) == bool:  # 最后一个参数：是否外连接
                    outer = join[-1]
                    j_args = join[:-1]
                else:
                    outer = False
                    j_args = join
                q = self.add_join(q, *j_args, outer=outer)
        if self.group_bys:
            for gb in self.group_bys:
                q = q.group_by(self.f(gb))
        if self.order_bys:
            for ob in self.order_bys:
                q = q.order_by(self.f(ob))

        return q

    def _add_kw_filters(self, q, kwargs):
        if not kwargs:
            return q
        for k, v in kwargs.items():
            if v is None:
                continue
            if is_collection(v):
                op, value = v  # Op , value
            else:
                op, value = Op.EQU, v
            q = self.add_filter(q, k, value, op)
        return q

    def query(self, make_q=None, convert_rec=None, post_process=None, distinct=False,*args, **kwargs) -> QueryR:
        """
        执行查询。调用者可传入3个回调函数，分别用于组装query、转换结果集的记录、对完整结果集的后处理.
        查询结果以DataFrame类型保存到_result属性中，可以使用result()方法获取。
        :param make_q: 用于组装query的回调函数，对应的签名为 (q:Query,qs:Selection,*args,**kwargs)->Query。其中：
            q - 待组装的query对象，应采用链式处理返回;
            ss - Selection对象自身，可用于处理反射类、反射属性等；
            args,kwargs - 外部调用者传递给q_maker函数的动态参数，通常用于查询条件
        :param convert_rec: 行记录转换回调函数，对应的签名为 (qs:Selection,r,index:int)->Any。其中：
            r - 行记录对象,query.all()返回的行元素
            index - 记录在结果集中的序号
        :param post_process: 后处理回调函数，对应的签名为 (qs:Selection,result:DataFrame)->Any。
            返回结果将被保存在result属性中，可以用result()方法多次获取。
        :param distinct: 执行查询时是否增加distinct()
        :param args:
        :param kwargs:
        :return:
        """

        def to_record(rec, index):
            if self.convert_record:  # 回调
                return self.convert_record(self, rec, index)
            else:
                return rec

        # 准备钩子函数
        if make_q:
            self.set_make_q(make_q)
        if convert_rec:
            self.set_convert_rec(convert_rec)
        if post_process:
            self.set_post_process(post_process)

        # 将参数放入上下文中
        if args:
            self.query_args.clear()
            self.query_args.extend(args)
        if kwargs:
            self.query_kwargs.clear()
            self.query_kwargs.update(kwargs)
        try:
            # 创建query对象
            q = self._create_query()
            # 回调 maker 得到完整的query
            if self.make_q:
                q = self.make_q(q, self, *args, **kwargs)
            else:
                q = self._add_kw_filters(q, kwargs)
            set_attr(self, 'sql', f"{q}")

            col_types = field_types(q)
            if distinct:
                q = q.distinct()
            # 获取查询结果并转换
            records = [to_record(result, index) for index, result in enumerate(q.all())]
        finally:
            self.session.close()

        dtypes = {col: 'Int64' for col in col_types if isinstance(col_types[col], INTEGER_TYPES)}
        df = pd.DataFrame(records)
        if not df.empty:
            df = df.astype(dtypes)

        # 后处理
        if self.post_process and records:  # 仅在有返回结果时调用
            df = self.post_process(self, df)
            if df is None:
                raise NotImplementedError('指定了后处理方法，但没有返回结果集', self.post_process)

        self._result = QueryR(df, self.column_spec())
        return self._result

    def dataset(self):
        return self._result


class MetaDb:
    DEFAULT_MODEL_TPL = 'db_models.ptl'
    DEFAULT_DATABASE_URI = 'sqlite:///:memory:'

    def __init__(self, database_url=None, echo=False):
        if database_url:
            self.database_url = database_url
        else:
            self.database_url = MetaDb.DEFAULT_DATABASE_URI
        self.engine = create_engine(self.database_url, echo=echo)
        self.meta = MetaData()
        self.meta.reflect(bind=self.engine, views=True)
        self.Base = automap_base(metadata=self.meta)
        self.Base.prepare()
        self._table_unique_cols = {}
        self._table_name_map = {t.name.upper(): t.name for t in self.meta.sorted_tables}  # 为了忽略大小写，建立大写表名与原始表名之间的映射
        self.map_tables()
        self.Session = sessionmaker(bind=self.engine)

    def map_tables(self):
        cls_names = set(self.Base.classes.keys())
        tab_names = set([t.name for t in self.meta.sorted_tables])
        diff_names = tab_names - cls_names
        if diff_names:
            print('WARNING:', f'表{diff_names}未能自动映射为model类，请仔细检查这些表是否定义了主键。')
        for table_name in self.Base.classes.keys():
            # 将 Automap 反射出的类其数据库实例默认为不可改变，将其转换为可更改实例值的ORM类
            self.Base.classes[table_name].__table__.info['autoload'] = True
        # 获得每个table的主键/唯一键/唯一索引
        for t in self.Base.metadata.sorted_tables:
            unique_cols = []
            if t.primary_key.columns:
                unique_cols.append([col.name for col in t.primary_key.columns])  # 主键
            for index in t.indexes:  # 唯一索引
                if index.unique:
                    unique_cols.append([col.name for col in index.columns])
            for constraint in t.constraints:  # 唯一约束
                if hasattr(constraint, 'columns') and constraint.__class__.__name__ == 'UniqueConstraint':
                    unique_cols.append([col.name for col in constraint.columns])
            self._table_unique_cols[t.name] = unique_cols

    def foreign_keys(self, tab_name):
        inspector = inspect(self.engine)
        return inspector.get_foreign_keys(self.m_table_name(tab_name))

    def m_table_name(self, tab_name: str):
        return self._table_name_map[tab_name.upper()]

    def m_class_name(self, tab_name: str):
        table_name = self.m_table_name(tab_name).capitalize()
        return f'_{table_name}'

    def tables(self):
        return self.meta.sorted_tables

    def table(self, table_name):
        # return self.meta.tables[self.m_table_name(table_name)]
        return CaseInsensitiveDict(self.meta.tables)[table_name]

    def table_unique_columns(self, table_name):
        # return self._table_unique_cols[self.m_table_name(table_name)]
        return CaseInsensitiveDict(self._table_unique_cols)[table_name]

    def model_class(self, table_name):
        # return self.Base.classes[self.m_table_name(table_name)]
        return CaseInsensitiveDict(self.Base.classes)[table_name]

    def _save_records(self, table_name, records, on_exists, on_without, pkey_columns=None):
        def single_record(r):
            record_in_db = None
            # 逐条规则检查
            for uq_cols in unique_rules:
                do_query = False
                q = session.query(cls)
                for key_col in uq_cols:  # 逐个字段拼筛选条件
                    if r.get(key_col):
                        fld = getattr(cls, key_col)
                        q = q.filter(fld == r[key_col])
                        do_query = True
                if do_query:
                    record_in_db = q.first()
                    if record_in_db:
                        break
            if record_in_db and on_exists:
                on_exists(session, r, record_in_db)
            elif on_without:
                on_without(session, r, record_in_db)

        session = self.Session()
        # 检查r在数据库中对应的的记录是否存在，如果存在则执行on_exists，否则执行on_without
        # 通过数据库表的唯一约束规则对应的字段（集合）作为查询条件判断是否存在
        unique_rules = self.table_unique_columns(table_name) if not pkey_columns else [pkey_columns]
        cls = self.model_class(table_name)

        # 逐条记录处理，批量提交
        for rec in records if isinstance(records, list) else [records]:
            single_record(rec)

        session.commit()
        session.close()

    def insert_or_update(self, table_name, records, pkey_columns=None):
        def copy_r(o, r):
            for key, value in r.items():
                setattr(o, key, value)
            return o

        def exec_exists(session, r, record_in_db):
            print(f'记录已经存在,UPDATE : {r}')
            copy_r(record_in_db, r)

        def exec_without(session, r, record_in_db):
            print(f'插入新记录: {r}')
            new_record = copy_r(self.model_class(table_name)(), r)
            session.add(new_record)

        self._save_records(table_name, records, exec_exists, exec_without, pkey_columns)

    def delete(self, table_name, records, pkey_columns=None):
        def exec_exists(session, r, record_in_db):
            print(f'找到并删除: {r}')
            session.delete(record_in_db)

        self._save_records(table_name, records, exec_exists, on_without=None, pkey_columns=pkey_columns)

    def generate_code(self, tpl_file=DEFAULT_MODEL_TPL, py_filename=None, save_to_file=False, name_pattern=None,
                      **kwargs):
        tables = []
        for name in self.Base.classes.keys():
            if not name_pattern:
                tables.append(self.table(name))
            elif re.match(name_pattern, name):
                tables.append(self.table(name))

        if not py_filename:
            py_filename = 'metadb_autogen.py'

        # 渲染模板
        context = {'meta': self,
                   'tables': tables,
                   'tpl_file': tpl_file,
                   'py_filename': py_filename,
                   'create_time': datetime.now(),
                   }
        context.update(kwargs)
        env = jinja2.Environment(loader=jinja2.FileSystemLoader("../../templates"))
        rendered_code = env.get_template(tpl_file).render(**context)
        # 保存到文件
        if save_to_file:
            with open(py_filename, 'w', encoding='utf-8') as file:
                file.write(rendered_code)
        return rendered_code

    def tables_summary(self):
        ret = []
        for table in self.tables():
            o_table = o2d(table, ['key', 'name', 'comment', 'description', ])
            o_table['columns'] = [
                o2d(c, ['key',
                        'name',
                        'comment',
                        'description',
                        '*type',
                        'primary_key',
                        'nullable',
                        '*autoincrement',
                        'unique',
                        ]) for c in table.columns]
            o_table['uniques'] = self.table_unique_columns(table.name)
            o_table['foreign_keys'] = self.foreign_keys(table.name)
            ret.append(o_table)
        return ret


class QueryManager:
    DEFAULT_NAME_SPACE = 'default'
    _instance = {}

    def __init__(self, name_space=DEFAULT_NAME_SPACE):
        self._name_space = name_space
        self._ds_dict = {}
        self._qmaker = {}

    def get_selection(self, ds_key) -> Selection:
        selection = self._ds_dict.get(ds_key)
        if isinstance(selection, str):
            selection = call_exp(selection)
            self._ds_dict[ds_key] = selection
        return selection

    def register_selection(self, ds_key: str, selection: str | Selection):
        # if isinstance(selection, str):
        #     selection = yut.call_exp(selection)
        self._ds_dict[ds_key] = selection

    def register_qmaker(self, ds_key: str, maker: Callable):
        self._qmaker[ds_key] = maker

    def get_qmaker(self, ds_key):
        return self._qmaker.get(ds_key)

    def name_space(self):
        return self._name_space

    def query_selection(self, ds_key=None, selection=None, condition: dict = None, convert_rec=None, post_process=None):

        def default_qmaker(q, qs: Selection, **kwargs):
            for c in qs.get_conditions():
                if c is not None:
                    q = qs.add_filter(q, qs.f(c.attr), c.value, c.op)
            ext_maker = self.get_qmaker(ds_key) if ds_key else None
            if ext_maker and is_execute_able(ext_maker):  # ds_key存在查询扩展编排器
                q = ext_maker(q, qs, qc=condition, **kwargs)
            return q

        if selection is None:
            selection = self.get_selection(ds_key)
        if selection is None:
            raise KeyError(f'无法获取Selection: ds_key={ds_key}')
        selection.push_conditions(condition, clear=True)
        qr: QueryR = selection.query(make_q=default_qmaker,
                                     convert_rec=convert_rec,
                                     post_process=post_process,
                                     ds_key=ds_key)
        return qr

    def selection_count(self):
        return len(self._ds_dict)

    @staticmethod
    def instance(name_space=DEFAULT_NAME_SPACE):
        fac = QueryManager._instance.get(name_space)
        if fac is None:
            fac = QueryManager(name_space)
            QueryManager._instance[name_space] = fac
        return fac



# -*- mes.py: python ; coding: utf-8 -*-
# 性能度量包
import time
import timeit
import cProfile
import pstats
from functools import wraps
from contextlib import contextmanager
from typing import Callable, Any, Optional


class CodeTimer:
    """
    一个用于测量代码执行时间的工具类，支持上下文管理器和装饰器两种使用方式
    """

    def __init__(self, name: str = None, print_report: bool = True):
        """
        初始化计时器

        :param name: 计时器名称，用于标识不同的计时段
        :param print_report: 是否在结束时自动打印报告
        """
        self.name = name or "Unnamed"
        self.print_report = print_report
        self.start_time = 0
        self.end_time = 0
        self.elapsed_time = 0

    def __enter__(self):
        """
        上下文管理器入口，开始计时
        """
        self.start_time = time.perf_counter()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        上下文管理器出口，结束计时并计算耗时
        """
        self.end_time = time.perf_counter()
        self.elapsed_time = self.end_time - self.start_time

        if self.print_report:
            print(f"{self.name} executed in {self.elapsed_time:.6f} seconds")

    def __call__(self, func):
        """
        使实例可作为装饰器使用
        """

        @wraps(func)
        def wrapped(*args, **kwargs):
            with self.__class__(name=func.__name__, print_report=self.print_report):
                return func(*args, **kwargs)

        return wrapped


def time_it(func: Callable = None, *, number: int = 1, repeat: int = 1) -> Any:
    """
    使用timeit模块测量函数执行时间的装饰器

    :param func: 被装饰的函数
    :param number: 每次测量中执行函数的次数
    :param repeat: 重复测量的次数
    :return: 函数执行结果和计时结果的元组 (result, timings)
    """
    if func is None:
        return lambda f: time_it(f, number=number, repeat=repeat)

    @wraps(func)
    def wrapper(*args, **kwargs):
        # 执行函数并获取结果
        result = func(*args, **kwargs)

        # 测量执行时间
        timer = timeit.Timer(lambda: func(*args, **kwargs))
        timings = timer.repeat(repeat=repeat, number=number)

        # 打印报告
        min_time = min(timings) / number
        avg_time = sum(timings) / len(timings) / number
        max_time = max(timings) / number

        print(f"\n--- {func.__name__} 执行时间报告 ---")
        print(f"执行次数: {number} 次/测量, 重复测量: {repeat} 次")
        print(f"最短单次执行时间: {min_time:.6f} 秒")
        print(f"平均单次执行时间: {avg_time:.6f} 秒")
        print(f"最长单次执行时间: {max_time:.6f} 秒")
        print(f"总执行时间: {sum(timings):.6f} 秒")

        return result, timings

    return wrapper


@contextmanager
def profile_code(sort_by: str = 'cumulative', limit: int = 10):
    """
    使用cProfile分析代码性能的上下文管理器

    :param sort_by: 排序依据，可选 'cumulative', 'time', 'calls' 等
    :param limit: 显示的行数限制
    """
    profiler = cProfile.Profile()
    profiler.enable()

    try:
        yield
    finally:
        profiler.disable()
        stats = pstats.Stats(profiler).sort_stats(sort_by)
        print("\n--- 性能分析报告 ---")
        stats.print_stats(limit)


def measure_performance(func: Callable = None, *,
                        enable_timeit: bool = True,
                        enable_profile: bool = False,
                        timeit_kwargs: Optional[dict] = None,
                        profile_kwargs: Optional[dict] = None):
    """
    综合性能测量装饰器，可以同时使用多种测量方法

    :param func: 被装饰的函数
    :param enable_timeit: 是否启用timeit测量
    :param enable_profile: 是否启用cProfile分析
    :param timeit_kwargs: 传递给time_it装饰器的参数
    :param profile_kwargs: 传递给profile_code的参数
    """
    if func is None:
        return lambda f: measure_performance(
            f,
            enable_timeit=enable_timeit,
            enable_profile=enable_profile,
            timeit_kwargs=timeit_kwargs,
            profile_kwargs=profile_kwargs
        )

    timeit_kwargs = timeit_kwargs or {}
    profile_kwargs = profile_kwargs or {}

    @wraps(func)
    def wrapper(*args, **kwargs):
        # 如果有性能分析，先处理
        if enable_profile:
            with profile_code(**profile_kwargs):
                if enable_timeit:
                    decorated_func = time_it(func, **timeit_kwargs)
                    return decorated_func(*args, **kwargs)
                else:
                    return func(*args, **kwargs)
        else:
            if enable_timeit:
                decorated_func = time_it(func, **timeit_kwargs)
                return decorated_func(*args, **kwargs)
            else:
                return func(*args, **kwargs)

    return wrapper


# 示例用法
if __name__ == "__main__":
    # 示例1: 使用CodeTimer类作为上下文管理器
    with CodeTimer("示例代码块1"):
        # 这里放你要测试的代码
        sum(range(1000000))


    # 示例2: 使用CodeTimer类作为装饰器
    @CodeTimer("计算平方和")
    def calculate_square_sum(n):
        return sum(i * i for i in range(n))


    calculate_square_sum(1000000)


    # 示例3: 使用time_it装饰器
    @time_it(number=10, repeat=3)
    def example_function(n):
        return sum(range(n))


    result, timings = example_function(1000000)
    print(f"计算结果: {result}")

    # 示例4: 使用profile_code上下文管理器
    with profile_code(sort_by='time', limit=5):
        sum(i * i for i in range(1000000))


    # 示例5: 使用综合性能测量装饰器
    @measure_performance(
        enable_timeit=True,
        enable_profile=True,
        timeit_kwargs={'number': 10, 'repeat': 3},
        profile_kwargs={'sort_by': 'time', 'limit': 5}
    )
    def performance_example(n):
        return sum(i * i * i for i in range(n))


    performance_example(10000)



# -*- ymodel.py: python ; coding: utf-8 -*-
#####################################################
# MVC机制下的Model类
#####################################################
import os
from typing import Callable, Any, List, Optional

import pandas as pd
from PySide6.QtCore import QAbstractTableModel, QModelIndex, Qt, QAbstractItemModel, QSize
from pandas import DataFrame

from . import TranslateDirection, filename_for_save, alignment_flag
from .yfield import FACTORY
from yus import ColumnSpec, Op, UnitType
from yut import is_collection, format_obj, get_attr, set_attr, is_execute_able


class PandasModel(QAbstractTableModel):

    def __init__(self,
                 data: DataFrame,
                 spec: list[ColumnSpec],
                 display_columns: list[str] = None,
                 custom_style_handler: Callable = None,
                 for_edit=False):
        """
        Pandas DataFrame 模型
        :param data: Pandas DataFrame
        :param spec: 列规格列表
        :param display_columns: 显示列名称
        :param custom_style_handler: 单元格格式处理函数，签名：(role:ForegroundRole|BackgroundRole|FontRole, index, col_name, cell_data)->Any
        :param for_edit: 模型的数据是否可以修改
        """
        super().__init__()
        self._data = data
        self._spec = {get_attr(o, 'name'): ColumnSpec.from_obj(o) for o in spec} if spec else {}
        self._display_columns = display_columns if display_columns is not None else self._data.columns.tolist()
        self.for_edit = for_edit
        self._custom_style_handler = custom_style_handler
        self._sort_order = [False] * len(self._display_columns)
        self._current_row_number = 0 if self.rowCount() > 0 else -1
        self._field_painters = {name: FACTORY.get_painter(spec) for name, spec in self._spec.items()}
        self._merged_cells = {}
        self._merged_columns = []
        self._cache = {}

    def set_display_columns(self, display_columns: list[str] = None):
        self._display_columns = display_columns if display_columns is not None else self._data.columns.tolist()

    def set_custom_style_handler(self, style_handler):
        self._custom_style_handler = style_handler

    def custom_style_handler(self):
        return self._custom_style_handler

    def update_data(self, data):
        self._data = data
        self._current_row_number = 0 if self.rowCount() > 0 else -1
        self._cache.clear()

    def load(self, data, spec: list):
        self._data = data
        self._spec = {sp.name: sp for sp in spec}
        self._current_row_number = 0 if self.rowCount() > 0 else -1
        self._cache.clear()

    def data_frame(self):
        return self._data

    def column_spec(self, col: int | str) -> ColumnSpec:
        if isinstance(col, int):
            keys_list = list(self._spec.keys())
            index = col if col >= 0 else len(keys_list) + col  # 可以指定负数，从尾部取索引
            return self._spec.get(keys_list[index])
        else:
            return self._spec[col]

    def column_specs(self):
        return self._spec.values()

    def rowCount(self, parent=QModelIndex()):
        return self._data.shape[0]

    def columnCount(self, parent=QModelIndex()):
        return len(self._display_columns)

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if not index.isValid():
            return None
        # 使用缓存
        cache_key = (index.row(), index.column(), role)
        if cache_key in self._cache:
            return self._cache[cache_key]
        ret = self._real_data(index, role)
        self._cache[cache_key] = ret
        return ret

    def _real_data(self, index, role):
        col_name = self._display_columns[index.column()]
        spec = self.column_spec(col_name)
        field_painter = self._field_painters.get(col_name)
        row_dat = self.data_of_row(index.row())
        cell_data = row_dat[col_name]  # self._data.iloc[index.row()][col_name] #

        if role == Qt.ItemDataRole.EditRole:
            return cell_data

        if role == Qt.ItemDataRole.UserRole:  # 返回当前行对象
            col_name = self._display_columns[index.column()]
            row_dat['_value'] = row_dat[col_name]
            return row_dat

        if role == Qt.ItemDataRole.DisplayRole:
            if field_painter:
                return field_painter.get_text(cell_data, row_dat)
            else:
                return '' if pd.isna(cell_data) or cell_data is None else str(cell_data)

        if role == Qt.ItemDataRole.ToolTipRole:  # 使用ToolTipRole处理链接跳转信息
            if spec.is_link():
                return get_attr(spec, 'link_to')

        if self._custom_style_handler:  # 如果定义了格式处理函数，调用并优先使用。
            ret = self._custom_style_handler(role, index, col_name, cell_data)
            if ret is not None:
                return ret

        if role == Qt.ItemDataRole.TextAlignmentRole:
            alignment = get_attr(spec, 'alignment')
            if alignment:  # 优先使用列规格中的对齐属性
                return alignment_flag(alignment)
            elif field_painter:
                fp_align = get_attr(field_painter, 'alignment')
                if fp_align:
                    return fp_align() if is_execute_able(fp_align) else alignment_flag(fp_align)
            else:
                return Qt.AlignmentFlag.AlignCenter

        return None

    def data_of_row(self, row_num):
        return self._data.iloc[row_num].to_dict()

    def current_row_number(self):
        return self._current_row_number

    def data_of_current_row(self):
        if self._current_row_number >= 0:
            return self.data_of_row(self._current_row_number)
        else:
            return None

    def set_current_row_number(self, row_number):
        if 0 <= row_number < self.rowCount():
            self._current_row_number = row_number

    def headerData(self, section, orientation, role=Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:  # 列标题
                col_name = self._display_columns[section]
                return self.column_spec(col_name).title
            else:  # 行标题
                # row_dat = self.data_of_row(section)
                # col = self._display_columns[0]
                return f"{section + 1:d}"
        else:
            return super().headerData(section, orientation, role)

    def setData(self, index, value, role=Qt.ItemDataRole.EditRole):
        if index.isValid() and role == Qt.ItemDataRole.EditRole:
            col_name = self._display_columns[index.column()]
            self._data.iat[index.row(), col_name] = value
            self.dataChanged.emit(index, index, [Qt.ItemDataRole.EditRole])
            return True
        return False

    def setForEdit(self, value):
        self.for_edit = value

    def isForEdit(self):
        return self.for_edit

    def flags(self, index):
        if self.isForEdit():
            return Qt.ItemFlag.ItemIsEditable | Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable
        else:
            return Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsEnabled

    def sort(self, column, order=...):
        self.layoutAboutToBeChanged.emit()
        order = self._sort_order[column]
        self._data = self._data.sort_values(by=self._display_columns[column], ascending=order)
        self._sort_order[column] = not self._sort_order[column]
        self._cache.clear()
        self.layoutChanged.emit()

    def updateMergedCells(self, columns_to_merge):
        self._merged_cells = {}
        self._merged_columns = columns_to_merge
        self._cache.clear()

        for col_name in columns_to_merge:
            if col_name not in self._display_columns:
                continue

            col_idx = self._display_columns.index(col_name)
            current_value = None
            start_row = 0

            for row in range(len(self._data)):
                value = self._data.iloc[row, self._data.columns.get_loc(col_name)]

                if value != current_value:
                    # 只有当有连续相同值时才合并
                    if current_value is not None and row > start_row:
                        span_rows = row - start_row
                        self._merged_cells[(start_row, col_idx)] = (span_rows, 1)
                    current_value = value
                    start_row = row

            # 处理最后一段
            if current_value is not None and len(self._data) > start_row:
                span_rows = len(self._data) - start_row
                if span_rows > 1:
                    self._merged_cells[(start_row, col_idx)] = (span_rows, 1)

    def span(self, index):
        return self._merged_cells.get((index.row(), index.column()), (1, 1))

    def summary(self, col_name):
        return 0.0 if self._data.empty else self._data[col_name].sum()

    def sort_order(self, column) -> bool:
        return self._sort_order[column]

    def export(self, filename=None, file_format='excel', select_file=False):
        def to_markdown(df_src, file_name):
            # 列标题
            df_src.rename(columns={spec.name: spec.title for spec in self.column_specs()}, inplace=True)
            df_src.to_markdown(file_name)

        df = self.df_for_display()
        ext, fn = {'excel': ('xlsx',
                             lambda f, columns, headers: df.to_excel(f, sheet_name='data',
                                                                     columns=columns,
                                                                     header=headers,
                                                                     engine='openpyxl')),
                   'csv': ('csv',
                           lambda f, columns, headers: df.to_csv(f, columns=columns, header=headers,
                                                                 encoding='utf-8')),
                   'html': ('html',
                            lambda f, columns, headers: df.to_html(f, columns=columns, header=headers, )),
                   'markdown': ('md',
                                lambda f, columns, headers: to_markdown(df, f)),
                   }[file_format]
        if not filename or select_file:
            path = filename if filename else os.curdir
            filename = filename_for_save(path=path, caption=f'请选择待保存的{file_format}文件位置和文件名称',
                                         file_filter=f"*.{ext}")
        elif not filename.endswith(ext):
            filename = f"{filename}.{ext}"
        exp_columns = self._display_columns
        exp_headers = [self.column_spec(c).title for c in exp_columns]
        if filename:
            fn(filename, exp_columns, exp_headers)
            return filename
        else:
            return None

    def df_for_display(self) -> DataFrame:
        # 处理Choose列，使用显示文字替换原始值
        df = self.data_frame().copy()
        column_specs = self.column_specs()
        choose_columns = [spec.name for spec in column_specs if spec.utype in (UnitType.CHOOSE, UnitType.LOOKUP)]
        for col in choose_columns:
            try:
                col_idx = self._display_columns.index(col)
                new_dat = [self.data(self.index(i, col_idx),
                                     Qt.ItemDataRole.DisplayRole) for i in range(self.rowCount())]
                df[col] = new_dat
            except ValueError:
                pass
        return df


class YTreeItem:
    """Y系列树节点基类"""

    def __init__(self, data: List[Any], parent=None):
        self.item_data = data
        self.parent_item = parent
        self.child_items = []
        self.has_children = False

    def appendChild(self, item) -> None:
        self.child_items.append(item)
        item.parent_item = self
        self.has_children = True

    def child(self, row: int) -> Optional['YTreeItem']:
        return self.child_items[row] if 0 <= row < len(self.child_items) else None

    def childCount(self) -> int:
        return len(self.child_items)

    def row(self) -> int:
        return self.parent_item.child_items.index(self) if self.parent_item else 0

    def data(self, column: int) -> Any:
        return self.item_data[column] if 0 <= column < len(self.item_data) else None

    def parent(self) -> Optional['YTreeItem']:
        return self.parent_item


class YTreeModel(QAbstractItemModel):
    """
    Y系列通用树模型
    """

    def __init__(self,
                 data: pd.DataFrame,
                 spec: list[ColumnSpec],
                 id_extractor: Callable[[Any], str],
                 parent_id_extractor: Callable[[Any], str],
                 caption_exp=None,
                 display_columns: Optional[List[str]] = None,
                 icon_provider: Optional[Callable[[YTreeItem], Any]] = None,
                 parent=None):
        super().__init__(parent)
        self._data = data
        self._spec = {get_attr(o, 'name'): ColumnSpec.from_obj(o) for o in spec}
        self._field_painters = {name: FACTORY.get_painter(spec) for name, spec in self._spec.items()}
        self.id_extractor = id_extractor
        self.parent_id_extractor = parent_id_extractor
        self.icon_provider = icon_provider
        self.caption_exp = caption_exp

        self.columns = display_columns if display_columns else list(data.columns)
        self.records = self._data.to_dict('records')
        self.root_item = YTreeItem([""] * len(self.columns))
        self.nodes = {}
        self._setup_model_data()

    def column_spec(self, col: int | str) -> ColumnSpec:
        if isinstance(col, int):
            keys_list = list(self._spec.keys())
            index = col if col >= 0 else len(keys_list) + col  # 可以指定负数，从尾部取索引
            return self._spec.get(keys_list[index])
        else:
            return self._spec[col]

    def column_specs(self):
        return self._spec.values()

    def _setup_model_data(self) -> None:
        # sorted_data = self._data.to_dict('records')
        sorted_data = sorted(self.records, key=lambda x: len(str(self.id_extractor(x))))
        self.nodes[''] = self.root_item

        for item in sorted_data:
            item_id = str(self.id_extractor(item))
            item_data = [str(item.get(col, "")) for col in self.columns]
            new_node = YTreeItem(item_data)
            set_attr(new_node, 'record', item)

            parent_id = str(self.parent_id_extractor(item))
            parent_node = self.nodes.get(parent_id, self.root_item)

            parent_node.appendChild(new_node)
            self.nodes[item_id] = new_node

        self._mark_has_children()

    def _mark_has_children(self) -> None:
        def process_item(item: YTreeItem) -> None:
            for i in range(item.childCount()):
                child = item.child(i)
                if child.childCount() > 0:
                    child.has_children = True
                process_item(child)

        process_item(self.root_item)

    def index(self, row: int, column: int, parent=QModelIndex()) -> QModelIndex:
        if not self.hasIndex(row, column, parent):
            return QModelIndex()
        parent_item = self._get_item(parent)
        child_item = parent_item.child(row)
        return self.createIndex(row, column, child_item) if child_item else QModelIndex()

    def parent(self, index: QModelIndex) -> QModelIndex:
        if not index.isValid():
            return QModelIndex()
        child_item = index.internalPointer()
        parent_item = child_item.parent()
        return QModelIndex() if parent_item == self.root_item else self.createIndex(parent_item.row(), 0, parent_item)

    def rowCount(self, parent=QModelIndex()) -> int:
        return self._get_item(parent).childCount()

    def columnCount(self, parent=QModelIndex()) -> int:
        return len(self.columns) + 1  # 多一个空列，确保最右侧的列宽度准确

    def data(self, index: QModelIndex, role=Qt.ItemDataRole.DisplayRole) -> Any:
        if not index.isValid():
            return None

        item = index.internalPointer()
        column = index.column()
        is_blank = column >= len(self.columns) or column < 0

        col_name = self.columns[column] if not is_blank else ''
        cell_data = item.data(column) if not is_blank else ''
        row_dat = {col: item.data(i) for i, col in enumerate(self.columns)} if not is_blank else {}
        field_painter = self._field_painters.get(col_name) if not is_blank else None

        if role == Qt.ItemDataRole.EditRole:
            return cell_data

        if role == Qt.ItemDataRole.DisplayRole:
            if field_painter:
                return field_painter.get_text(cell_data, row_dat)
            else:
                return None if pd.isna(cell_data) else str(cell_data)

        if role == Qt.ItemDataRole.UserRole:
            return row_dat

        if role == Qt.ItemDataRole.DecorationRole and column == 0 and self.icon_provider:
            return self.icon_provider(item)

        if role == Qt.ItemDataRole.TextAlignmentRole and not is_blank:
            spec = self.column_spec(col_name)
            alignment = get_attr(spec, 'alignment')
            if alignment:
                return alignment_flag(alignment)
            if field_painter:
                return field_painter.alignment()

        if role == Qt.ItemDataRole.SizeHintRole:
            return QSize(-1, 24)

        return None

    def headerData(self, section: int, orientation: Qt.Orientation, role=Qt.ItemDataRole.DisplayRole) -> Any:
        if role == Qt.ItemDataRole.DisplayRole and orientation == Qt.Orientation.Horizontal:
            col = self.columns[section] if section < len(self.columns) else None
            return self.column_spec(col).title if col else None
        return None

    def _get_item(self, index: QModelIndex) -> YTreeItem:
        return index.internalPointer() if index.isValid() else self.root_item

    def flags(self, index):
        return Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable

    def data_frame(self):
        return self._data


def add_qc_value(o, attr, op: Op, value):
    v = (op.value, value)
    if isinstance(o, dict):
        o[attr] = v
    else:
        setattr(o, attr, v)
    return o


def translate_olist(direction: TranslateDirection, lookup, v_from, olist):
    """
    翻译对象属性
    :param direction: 翻译的方向: 正向 lookup[0] -> lookup[1]; 反向 lookup[1] -> lookup[0]
    :param lookup: 属性映射表['{CODE}','{CODE} - {NAME}']
    :param v_from: 待转换的属性值，可以是集合，如['0001','0002','0003']
    :param olist: 对象集合
    :return: 翻译后的结果值
    例如，有码表集合[
    Record(Code=100,Name='001',Text='Text1'),
    Record(Code=102,Name='002',Text='Text2'),
    Record(Code=103,Name='003',Text='Text3'),
    ]
    可以执行
        texts=translate_olist(TranslateDirection.TO_TEXT,['{Code}','{Code} - {Text}'],[100,102])
        返回['100 - Text1','102 - Text2']
    """
    v_to = []
    v_src = []
    lookup_from, lookup_to = (lookup[0], lookup[1]) if direction == TranslateDirection.TO_TEXT else (
        lookup[1], lookup[0])
    v_src.extend(v_from) if is_collection(v_from) else v_src.append(v_from)
    for vf in v_src:
        for o in olist:
            o_from = format_obj(o, lookup_from)
            if vf is not None and o_from.strip() == vf.strip():
                v_to.append(format_obj(o, lookup_to))
                break
    return ','.join(v_to)


def col_spec(column_specs: list, col_key, **kwargs):
    """
    依据列规格列表中的现有对象生成新的的列规格
    :param column_specs: 列规格列表
    :param col_key: 列名/列标识
    :param kwargs: 补充或替代的规格属性
    :return:
    """
    cp = [c for c in column_specs if col_key == c['name']]
    if len(cp) == 1:
        d = dict(cp[0])
        d.update(kwargs)
        return ColumnSpec(**d)
    else:
        return None



# -*- yfield.py: python ; coding: utf-8 -*-
#####################################################
# 字段及其控制器
#####################################################
import datetime
import re
from abc import abstractmethod

import pandas as pd
from PySide6.QtCore import Qt, QDate, Signal, QObject, QTime, QDateTime, QRect
from PySide6.QtGui import QDoubleValidator, QColor, QPainter, QBrush, QPalette
from PySide6.QtWidgets import QWidget, QFormLayout, QLineEdit, QLayout, \
    QSpinBox, QDateEdit, QLabel, QComboBox, QButtonGroup, QRadioButton, QCheckBox, QDateTimeEdit, \
    QStyledItemDelegate, QTableView, QSizePolicy, QTextEdit, QGroupBox, QVBoxLayout, QTreeView
from pandas import Timestamp

from yus import ColumnSpec, Op, UnitType, ChooseMode, DisplayAttitude
from yut import to_float, to_date, to_datetime, is_execute_able, format_date, has_attr, get_attr, \
    set_attr, format_datetime, o2d, is_collection, copy_obj, create_instance
from . import Stage, WorkMode, add_to_layout, img, current_theme, alignment_flag
from . import add_field, create_layout_widget, create_layout, clear_layout, set_label_color, load_icon, create_button
from .wdg import YearMonthSpin
from .ydict import DictFetcher


class YFieldWidget(QTextEdit):
    """
    数据字段只读显示组件
    """

    def __init__(self, parent, col_spec: ColumnSpec, value=None, bind_obj=None):
        super().__init__(parent=parent)
        self.factory = FACTORY
        self.col_spec = col_spec
        copy_obj(self, self.col_spec)
        self._value = value
        self._bind_obj = bind_obj
        self._text = str(self._value) if self._value else ''
        self._field_painter = self.factory.get_painter(self.col_spec)
        if self._bind_obj:
            self.bind(self._bind_obj)
        self.setup_ui()

    def setup_ui(self):
        self.setReadOnly(True)
        font = self.font()
        color = get_attr(self, 'color', None)
        if color:
            self.setTextColor(color)
        font_size = get_attr(self.col_spec, 'font_size', None)
        if font_size:
            font.setPointSize(font_size)
            self.setFont(font)
        font_bold = get_attr(self.col_spec, 'font_bold', None)
        if font_bold:
            font.setBold(font_bold)
            self.setFont(font)
        height = get_attr(self.col_spec, 'height', None)
        if self.col_spec.utype != UnitType.RICH_TEXT:
            self.setFixedHeight(self.fontMetrics().height() + 12)
            self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        else:  # 长文本类型，尽可能大
            self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        if height:
            self.setFixedHeight(height)  # self.fontMetrics().height())

    def field_alignment(self):
        spec = self.col_spec
        alignment = get_attr(spec, 'alignment')
        if alignment:  # 优先使用列规格中的对齐属性
            return alignment_flag(alignment)
        elif spec.utype in (UnitType.CURRENCY, UnitType.INT, UnitType.NUMBER):
            return Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter
        elif spec.utype in (UnitType.CHOOSE, UnitType.LOOKUP, UnitType.DATE):
            return Qt.AlignmentFlag.AlignCenter | Qt.AlignmentFlag.AlignVCenter
        elif spec.utype == UnitType.RICH_TEXT:
            return Qt.AlignmentFlag.AlignLeft
        return Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter

    def set_value(self, value):
        self._value = value
        self._text = self._field_painter.get_text(self._value, self._bind_obj)
        self.setUpdatesEnabled(False)
        self.setText(self._text)
        self.setUpdatesEnabled(True)

    def bind(self, obj):
        self._bind_obj = obj
        self.set_value(get_attr(self._bind_obj, self.col_spec.name))

    def add_to(self, layout):
        label = self.get_field_label() + ':'
        if isinstance(layout, QFormLayout):  # Form布局，label为widget部件，否则label为字符串
            add_field(layout, self, QLabel(label))
        else:
            add_field(layout, self, label)

    def get_field_label(self):
        # label 依次使用 title/comment/field属性
        label: str = get_attr(self, 'title', None)
        if label is None:
            label = get_attr(self, 'comment', None)
        if label is None:
            label = self.col_spec.name.upper()
        return label

    def setup_palette(self):
        spec = self.col_spec
        self.setAlignment(self.field_alignment())
        pl = self.palette()
        fp = self._field_painter
        if spec.is_link():  # 处理链接样式
            font = self.font()
            font.setUnderline(True)
            self.setFont(font)
            brush = self.palette().brush(QPalette.ColorRole.Link)
            pl.setColor(QPalette.ColorRole.Text, brush.color())
        elif hasattr(fp, 'get_color'):  # 调用fp的设置painter方法
            color = fp.get_color(self._value, self._text)
            if color:
                pl.setColor(QPalette.ColorRole.Text, color)
        self.setPalette(pl)

    def paintEvent(self, event):
        if self.col_spec.utype != UnitType.RICH_TEXT:  # 富文本不要调用setup_palette，否则会卡顿
            self.setup_palette()
        # 调用FieldPainter，自定义绘制
        if hasattr(self._field_painter, 'draw_field'):
            painter = QPainter(self.viewport())
            self._field_painter.draw_field(painter, self.rect(), self._value, self._text)
        else:
            super().paintEvent(event)


class YDisplayDelegate(QStyledItemDelegate):
    """
    只读显示列代理控制器。
    """
    MARGIN_LEFT = 4
    MARGIN_RIGHT = 4

    def __init__(self, view: QTableView | QTreeView, col_spec: ColumnSpec, **kwargs):
        super().__init__(parent=view)
        self.factory = FACTORY
        self.view = view
        self.col_spec = col_spec
        self.name = col_spec.name
        self.label = col_spec.title
        self.utype = self.factory.deduce_unit_type(col_spec=col_spec)  # UnitType(col_spec.unit_type())
        self.comment = col_spec.comment
        self.fv_kwargs = kwargs
        self._field_painter = self.factory.get_painter(self.col_spec, **self.fv_kwargs)

    def createEditor(self, parent, option, index):
        return QLineEdit(parent=parent)  # 也要返回一个QLineEdit，用于选中显示文字，复制粘贴

    def setEditorData(self, editor, index):
        model = index.model()
        value = index.model().data(index, Qt.ItemDataRole.EditRole)
        editor.setText(self._field_painter.get_text(value, model.data_of_row(index.row())))

    def setModelData(self, editor, model, index):
        pass  # 只读，不修改

    def setup_painter(self, painter, value, text):
        if hasattr(self._field_painter, 'setup_painter'):  # 调用fp的设置painter方法
            self._field_painter.setup_painter(painter, value, text)

    def setup_link_painter(self, painter):
        font = painter.font()
        font.setUnderline(True)
        painter.setFont(font)
        brush = self.view.palette().brush(QPalette.ColorRole.Link)
        painter.setFont(font)
        painter.setPen(brush.color())

    def paint(self, painter, option, index):
        model = index.model()
        fp = self._field_painter
        if model is None or not index.isValid():
            return
        value = model.data(index, Qt.ItemDataRole.EditRole)
        text = model.data(index, Qt.ItemDataRole.DisplayRole)
        link_to = model.data(index, Qt.ItemDataRole.ToolTipRole)
        painter.save()
        self.initStyleOption(option, index)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        # 设置painter
        if link_to:  # 处理链接样式
            self.setup_link_painter(painter)
        self.setup_painter(painter, value, text)
        # 调用FieldPainter，自定义绘制
        if hasattr(fp, 'draw_field'):
            fp.draw_field(painter, option.rect, value, text)
        elif text:  # 显示文本,None或空时不显示
            rect: QRect = option.rect.adjusted(self.MARGIN_LEFT, 0, 0 - self.MARGIN_RIGHT, 0)
            alignment = model.data(index, Qt.ItemDataRole.TextAlignmentRole)
            font = model.data(index, Qt.ItemDataRole.FontRole)
            fr_color = model.data(index, Qt.ItemDataRole.ForegroundRole)
            bg_color = model.data(index, Qt.ItemDataRole.BackgroundRole)
            if font:
                painter.setFont(font)
            if fr_color:
                painter.setPen(fr_color)
            if bg_color:
                painter.setBackground(bg_color)
            painter.drawText(rect, alignment, text)
        painter.restore()


class YEditorDelegate(YDisplayDelegate):
    """
    可编辑列代理控制器。
    """

    def createEditor(self, parent, option, index):
        field_editor = self.factory.get_editor(parent=parent,
                                               col_spec=self.col_spec,
                                               stage=Stage.IN_TABLE,
                                               **self.fv_kwargs)
        return field_editor

    def setEditorData(self, editor, index):
        value = index.data(Qt.ItemDataRole.EditRole)
        # print(f'[YField] - YEditorDelegate.setEditorData: {index.row(), index.column()} -> ', '%r' % value,
        #       f", editor = {editor}")
        if value is pd.NaT:
            value = None
        if value is not None:
            editor.set_value(value)

    def setModelData(self, editor, model, index):
        value = editor.get_value()
        # old_value = model.data(index, Qt.ItemDataRole.EditRole)
        # print('[YField] - YEditorDelegate.setModelData: %r' % value, f' -> {index.row(), index.column()}',
        #       ', old value = %r' % old_value)
        model.setData(index, value, Qt.ItemDataRole.EditRole)

    def paint(self, painter, option, index):
        if not self.view.indexWidget(index):  # 仅在没有活动的编辑器时调用父类的paint方法绘制默认文本显示内容。
            super().paint(painter, option, index)


#############################################################
# 字段编辑组件
#############################################################
class YFieldEditor(QWidget):
    """
    各种数据字段的录入/展示控制器
    适应一个model属性绑定多个element的场景。on_change -> 当前改动的内容传播到model，并更新所有element的显示内容
    需要创建一个信号，各Element值更改后发射此信号，form得到属性值后循环调用全部element的set_value
    """
    valueChanged = Signal(str)  # field

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent=parent)
        # 从初始化参数中解析属性
        copy_obj(self, kwargs)
        self.stage = get_attr(self, 'stage', Stage.IN_FORM)
        self.field: str = get_attr(self, 'field', None)
        if self.field is None:
            self.field = get_attr(self, 'name', None)
        # label 依次使用 label/comment/field属性
        self.label: str = get_attr(self, 'label', None)
        if self.label is None:
            self.label = get_attr(self, 'comment', None)
        if self.label is None:
            self.label = self.field.upper()
        self.desc = get_attr(self, 'desc')
        self.mandatory = get_attr(self, 'mandatory', False)
        self.read_only = get_attr(self, 'read_only', False)
        self.work_mode = get_attr(self, 'work_mode', WorkMode.EDIT)
        if self.label.startswith('*'):
            self.mandatory = True
            self.label = self.label[1:]
        # 准备内部属性
        self._binds = []
        self._label_widget = None
        # 初始化UI
        self.setLayout(create_layout('H'))
        self.layout().setContentsMargins(2, 0, 2, 0)
        # 设置初始值
        v = get_attr(self, 'value')
        if v is not None:
            self.set_value(v)

    def editor(self):
        # return self.findChild(QWidget, 'editor')
        return get_attr(self, '_editor')

    def tag_editor(self, editor: QWidget):
        # editor.setObjectName('editor')
        set_attr(self, '_editor', editor)

    def label_widget(self):
        if self._label_widget is None:
            self._label_widget = self.create_label_widget()
            self._label_widget.field_view = self
        return self._label_widget

    def bind(self, o, attr=None, only_has_attr=False):
        o_attr = attr if attr else self.field
        self._binds.append([o, o_attr, only_has_attr])
        if has_attr(o, o_attr):
            self.set_value(get_attr(o, o_attr))

    def unbind(self, o):
        for i in range(len(self._binds)):
            if o in self._binds[i]:
                self._binds.pop(i)

    def on_change(self):
        # 更新绑定对象的值
        v = self.get_value()
        for o, attr, only_has_attr in self._binds:
            if not only_has_attr or has_attr(o, attr):
                set_attr(o, attr, v)
        self.valueChanged.emit(self.field)

    def add_to(self, layout: QLayout, index: int = -1):
        """
        将控件加入到指定的布局中
        :param layout: 待加入的布局对象
        :return:
        """
        if isinstance(layout, QFormLayout):  # Form布局，label为widget部件，否则label为字符串
            add_field(layout, self, self.label_widget(), index)
        else:
            add_field(layout, self, f'* {self.label}' if self.mandatory else f'{self.label}', index)

    def create_label_widget(self) -> QWidget:
        w_label = create_layout_widget(QWidget, parent=self.parentWidget())
        if self.mandatory:
            mark = QLabel('*')
            set_label_color(mark, fore_color=current_theme.icon_color())
            w_label.layout().addWidget(mark, 0)
        w_label.layout().addWidget(QLabel(f"{self.label}:" if self.label else ""), 1)
        return w_label

    @abstractmethod
    def set_value(self, value):
        pass

    @abstractmethod
    def get_value(self):
        return None


class YFdText(YFieldEditor):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self.none_text = ''
        editor = QLineEdit(parent=self)
        editor.setReadOnly(self.read_only)
        # editor.setEnabled(not self.read_only)
        editor.setPlaceholderText(self.desc)
        editor.editingFinished.connect(self.on_change)
        self.layout().addWidget(editor)
        self.tag_editor(editor)

    def set_value(self, value):
        self.editor().setText(self.none_text if value is None else f'{value}')

    def get_value(self):
        return self.editor().text().strip()


class YFdRichText(YFieldEditor):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self.none_text = ''
        self.text_format = get_attr(self, 'text_format', 'plain')

        editor = QTextEdit(parent=self)
        editor.setReadOnly(self.read_only)
        editor.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        # editor.setEnabled(not self.read_only)
        editor.setPlaceholderText(self.desc)
        # editor.setStyleSheet("background-color: gray;")
        self.layout().addWidget(editor)
        self.tag_editor(editor)

    def set_value(self, value):
        self.editor().setHtml(self.none_text if value is None else f'{value}')

    def get_value(self):
        editor: QTextEdit = self.editor()
        if self.text_format == 'plain':
            return editor.toPlainText()
        elif self.text_format == 'markdown':
            return editor.toMarkdown()
        else:
            return editor.toHtml()


class YFdNumber(YFieldEditor):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self.none_text = ''
        self._value = 0.00
        self.decimals = get_attr(self, 'decimals', 2)
        editor = NumberEdit(parent=self)
        editor.setAlignment(Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignRight)
        editor.setValidator(QDoubleValidator(decimals=self.decimals, parent=self))
        editor.setReadOnly(self.read_only)
        # editor.setEnabled(not self.read_only)
        editor.editingFinished.connect(self.edit_finished)
        self.layout().addWidget(editor)
        self.tag_editor(editor)
        self._disp_value()

    def set_value(self, value):
        self._value = to_float(value) if value is not None else 0.0
        self._disp_value()

    def get_value(self):
        self._value = to_float(self.editor().text())
        return self._value

    def edit_finished(self):
        self.get_value()  # 为了更新self._value
        self._disp_value()
        self.on_change()

    def _disp_value(self):
        self.editor().setText(f'{self._value:.2f}')


class YFdInt(YFieldEditor):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self.min_value = get_attr(self, 'min_value', None)
        self.max_value = get_attr(self, 'max_value', None)
        editor = QSpinBox(parent=parent)
        editor.setRange(0, 2147483647)
        if self.min_value is not None:
            editor.setMinimum(self.min_value)
        if self.max_value is not None:
            editor.setMaximum(self.max_value)
        editor.setReadOnly(self.read_only)
        editor.setAlignment(Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignRight)
        # editor.setEnabled(not self.read_only)
        editor.setValue(0)
        editor.editingFinished.connect(self.on_change)
        self.layout().addWidget(editor)
        self.tag_editor(editor)

    def set_value(self, value):
        editor = self.editor()
        if value:
            editor.setValue(editor.valueFromText(f'{value}'))  # 整形、字符型、浮点型都可兼容

    def get_value(self):
        return self.editor().value()


class YFdSwitch(YFieldEditor):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self.disp_text = get_attr(self, 'disp_text', ['[NO]', '[YES]'])
        editor = QCheckBox(parent=parent)
        editor.setEnabled(not self.read_only)
        editor.setText(self.disp_text[1])
        editor.setChecked(False)
        self.layout().addWidget(editor)
        self.layout().setAlignment(Qt.AlignmentFlag.AlignCenter | Qt.AlignmentFlag.AlignVCenter)
        self.tag_editor(editor)

    def set_value(self, value):
        value = to_float(value)
        editor = self.editor()
        editor.setChecked(value > 0.01)

    def get_value(self):
        return 1 if self.editor().isChecked() else 0


# 金额部件
class YFdCurrency(YFdNumber):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self.decimals = 2

    def _disp_value(self):
        if self.stage in (Stage.IN_FORM, Stage.IN_QUERY):
            self.editor().setText(f'{self._value:,.2f}')  # 在Form/Query中，千分位
        else:
            self.editor().setText(f'{self._value:.2f}')  # 在表格中编辑，编辑组件不做千分位，外部的表格会通过get_text()千分位


class YFdYear(YFdInt):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self.min_value = get_attr(self, 'min_value', 2000)
        self.max_value = get_attr(self, 'max_value', 2999)
        self.editor().setRange(2000, 2099)
        self.editor().valueChanged.connect(self.on_change)
        self.set_value(datetime.date.today().year)


class YFdYearMonth(YFieldEditor):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        editor = YearMonthSpin(datetime.date.today().year, datetime.date.today().month)
        editor.valueChanged.connect(self.on_change)
        self.tag_editor(editor)

    def set_value(self, value):
        if value is not None:
            self.editor().set_monthcode(value)

    def get_value(self):
        return self.editor().month_code()


class YFdDateTime(YFieldEditor):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self.display_format = get_attr(self, 'display_format', 'yyyy-MM-dd HH:mm:ss.zzz (dddd)')
        self.calendar_popup = get_attr(self, 'calendar_popup', True)
        editor = QDateTimeEdit(QDateTime.currentDateTime(), parent=parent)
        editor.setDisplayFormat(self.display_format)
        editor.setReadOnly(self.read_only)
        # editor.setEnabled(not self.read_only)
        editor.setAlignment(Qt.AlignmentFlag.AlignCenter)
        editor.setMaximumDate(QDate(2999, 12, 31))
        editor.setMinimumDate(QDate(1900, 1, 1))
        editor.setCalendarPopup(self.calendar_popup)
        editor.editingFinished.connect(self.on_change)
        self.layout().addWidget(editor)
        self.layout().addStretch(0)
        self.tag_editor(editor)

    def set_value(self, value):
        if value is None:
            return
        dt = to_datetime(value)
        self.editor.setDate(QDate(dt.year, dt.month, dt.day))
        self.editor.setTime(QTime(dt.hour, dt.minute, dt.second, dt.microsecond))

    def get_value(self):
        d = self.editor.dateTime()
        return format_datetime(d.toPython()) if d else None


class YFdDate(YFieldEditor):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        editor = QDateEdit(parent=parent)
        editor.setReadOnly(self.read_only)
        # editor.setEnabled(not self.read_only)
        editor.setAlignment(Qt.AlignmentFlag.AlignCenter)
        editor.setMaximumDate(QDate(2999, 12, 31))
        editor.setMinimumDate(QDate(1900, 1, 1))
        editor.setCalendarPopup(True)
        today = datetime.date.today()
        editor.setDate(QDate(today.year, today.month, today.day))
        editor.editingFinished.connect(self.on_change)
        self.layout().addWidget(editor)
        self.layout().addStretch(0)
        self.tag_editor(editor)

    def set_value(self, value):
        if value is None:
            return
        date = to_date(value)
        self.editor().setDate(QDate(date.year, date.month, date.day))

    def get_value(self):
        d = self.editor().date()
        return format_date(d.toPython()) if d else None


class YFdChoose(YFieldEditor):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self.code_dict = DictFetcher.instance().get_dict(self)
        if self.code_dict is None:
            self.code_dict = {}
        self.query_op = get_attr(self, 'op', Op.IN)
        self.mode = get_attr(self, 'mode', ChooseMode.LIST)
        self.format = get_attr(self, 'format', '[{code}] - {name}')
        # 如果是查询场景，mode自动设置为checkbox,布局根据可选项数量确定为'V'或'G2'
        if self.stage == Stage.IN_QUERY and self.query_op != Op.EQU:
            self.mode = ChooseMode.CHECK
            self.layout_flag = get_attr(self, 'layout_flag', 'V') if len(self.code_dict.keys()) < 6 else 'G2'
        else:
            self.layout_flag = get_attr(self, 'layout_flag', 'H')
        # 设置UI
        editor = self.create_editor(parent)
        self.layout().addWidget(editor)
        self.tag_editor(editor)

    def create_editor(self, parent) -> None | QWidget | QComboBox:
        if ChooseMode.LIST == self.mode:  # 使用 ComboBox
            editor = QComboBox(parent=parent)
            editor.setEditable(False)
            for code, name in self.code_dict.items():
                editor.addItem(load_icon(img.P_LABEL), eval(f'f"{self.format}"'), code)
            editor.currentIndexChanged.connect(self.on_change)
            editor.setEnabled(not self.read_only)
            return editor

        elif ChooseMode.RADIO == self.mode:  # 使用ButtonGroup -> RadioButton
            editor = create_layout_widget(QWidget, parent=parent, layout_flag=self.layout_flag)
            editor.button_group = QButtonGroup(parent=parent)
            for code, name in self.code_dict.items():
                btn = QRadioButton(eval(f'f"{self.format}"'))
                btn.data = code
                add_to_layout(editor.layout(), btn)
                editor.button_group.addButton(btn)
            editor.setEnabled(not self.read_only)
            editor.button_group.buttonClicked.connect(self.on_change)
            return editor

        elif ChooseMode.CHECK == self.mode:  # 使用GroupBox -> CheckBox
            editor = create_layout_widget(QWidget, parent=parent, layout_flag=self.layout_flag)
            for code, name in self.code_dict.items():
                btn = QCheckBox(eval(f'f"{self.format}"'))
                btn.data = code
                add_to_layout(editor.layout(), btn)
                btn.stateChanged.connect(self.on_change)
            editor.setEnabled(not self.read_only)
            return editor
        else:
            return None

    def set_value(self, value):
        if value is None:
            return
        editor = self.editor()
        if ChooseMode.LIST == self.mode:  # 使用 ComboBox
            for index in range(editor.count()):
                if str(value) == str(editor.itemData(index, Qt.ItemDataRole.UserRole)):
                    editor.setCurrentIndex(index)
        elif ChooseMode.RADIO == self.mode:  # 使用ButtonGroup -> RadioButton
            for btn in editor.button_group.buttons():
                btn.setChecked(str(value) == str(btn.data))
        elif ChooseMode.CHECK == self.mode:  # 使用GroupBox -> CheckBox
            for btn in editor.children():
                if hasattr(btn, 'data'):
                    btn.setChecked(str(btn.data) in [str(v) for v in value])

    def get_value(self):
        editor = self.editor()
        if ChooseMode.LIST == self.mode:  # 使用 ComboBox
            return editor.currentData(Qt.ItemDataRole.UserRole)
        elif ChooseMode.RADIO == self.mode:  # 使用ButtonGroup -> RadioButton
            choose = editor.button_group.checkedButton()
            if choose:
                return choose.data
        elif ChooseMode.CHECK == self.mode:  # 使用GroupBox -> CheckBox
            values = []
            grp = editor
            if hasattr(grp, 'check_all'):
                if grp.check_all.isChecked():
                    return None
            for chd in grp.children():
                if hasattr(chd, 'data'):
                    if chd.isChecked():
                        values.append(chd.data)
            return values
        return None


class YFdLookup(YFieldEditor):
    """
    弹框查选组件。
    组件由文本框+点选按钮构成。点选按钮可唤出弹框，并从候选列表中选取记录。候选列表支持查询，支持多选。
    """

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self.lookup_key = get_attr(self, 'lookup_key')
        self.lookup = FACTORY.get_lookup(**kwargs) if self.lookup_key else None
        self.lookup_value = None
        self.lookup_display = ''
        editor = QLineEdit(self.lookup_display, parent=parent)
        self.lookup_btn = create_button('',
                                        trigger=self.on_select,
                                        flat=False,
                                        tip_text='从弹出窗口选择',
                                        icon_file=img.P_3DOTS_H)
        self.lookup_btn.setEnabled(not self.read_only)
        self.layout().addWidget(editor)
        self.layout().addWidget(self.lookup_btn)
        self.tag_editor(editor)

    def on_select(self):
        # btn = self.sender()
        if self.lookup is None:
            return
        flag, value, text = self.lookup.execute_lookup()
        if flag > 0:  # 用户执行了选择
            self.lookup_value = value
            self.lookup_display = text
            self.editor().setText(self.lookup_display)

    def set_value(self, value):
        self.lookup_value = value
        self.lookup_display = f'{value}' if value is not None else ''
        if self.lookup:
            self.lookup_display = self.lookup.to_text(value)
        self.editor().setText(self.lookup_display)

    def get_value(self):
        return self.lookup_value


class NumberEdit(QLineEdit):
    def __init__(self, parent=None):
        super(NumberEdit, self).__init__(parent=parent)

    def focusInEvent(self, e):
        if hasattr(self, 'amount'):  # 不显示千分位
            self.setText(f"{self.amount:.2f}")
        self.selectAll()


#############################################################
# 字段显示组件
#############################################################

class YFieldPainter(QObject):
    def __init__(self, **kwargs):
        super().__init__(parent=None)
        copy_obj(self, kwargs)

    def setup_painter(self, painter, value, text):
        pass

    @abstractmethod
    def get_text(self, value, record):
        return f"{value}"

    @staticmethod
    def get_attitude_color(attitude: DisplayAttitude):
        return {
            DisplayAttitude.POSITIVE: QColor(0x00, 0xD0, 0xA0),
            DisplayAttitude.WEAK_POSITIVE: QColor(0x60, 0xD0, 0xD0),
            DisplayAttitude.NEUTRAL: QColor(Qt.GlobalColor.lightGray),
            DisplayAttitude.WEAK_NEGATIVE: QColor(0xD0, 0xA0, 0x60),
            DisplayAttitude.NEGATIVE: QColor(0xD0, 0x60, 0x60),
            DisplayAttitude.INACTIVE: QColor(Qt.GlobalColor.darkGray),
        }.get(attitude)

    @staticmethod
    def alignment():
        return Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter


class YFpText(YFieldPainter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        color_match: dict = get_attr(self, 'color_match', None)
        display = get_attr(self, 'display', None)
        self._color_match = color_match
        self._display = display

    def get_color(self, value, text):
        if not self._color_match:
            return None
        # 根据显示文本设置颜色
        for pattern, color in self._color_match.items():
            if re.match(pattern, text):
                return QColor(color)
        return None

    def get_text(self, value, record):
        if value is None:
            return ''
        if self._display is None:
            return super().get_text(value, record)
        if is_execute_able(self._display):
            return self._display(value, record)
        l_ctx = o2d(record)
        l_ctx['value'] = value  # 将value也加入上下文，方便直接使用
        return eval(f'f"{self._display}"', __locals=l_ctx)

    def setup_painter(self, painter, value, text):
        color = self.get_color(value, text)
        if color:
            painter.setPen(QColor(color))


class YFpInt(YFieldPainter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.leading_zero = get_attr(self, 'leading_zero', 0)
        self.format = get_attr(self, 'format', '%d')
        self.zero = get_attr(self, 'zero', '0')

    def get_text(self, value, record):
        fmt = f"%0{self.leading_zero}d" if self.leading_zero else self.format
        v = None if pd.isna(value) else int(value)
        txt = fmt % v if v is not None else None
        if txt == '0':
            return get_attr(self, 'zero', txt)
        return txt

    @staticmethod
    def alignment():
        return Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter


class YFpSwitch(YFieldPainter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.disp_text = get_attr(self, 'disp_text', ['[NO]', '[YES]'])
        self.attitude_color = get_attr(self, 'attitude_color')

    def setup_painter(self, painter, value, text):
        color = self.get_color(value, text)
        if color:
            painter.setPen(color)

    def get_color(self, value, text):
        if not self.attitude_color:
            return None
        # 根据值设置颜色，开关值为ON时显示 WEAK_POSITIVE 颜色
        att = DisplayAttitude.INACTIVE
        if text == self.disp_text[1]:
            att = DisplayAttitude.WEAK_POSITIVE
        return self.get_attitude_color(att)

    def get_text(self, value, record):
        if value is None:
            return self.disp_text[0]
        value = str(value).strip().upper()
        p = r"(1.*)|(T.*)|(Y.*)"
        if re.match(p, value):
            return self.disp_text[1]
        else:
            return self.disp_text[0]

    @staticmethod
    def alignment():
        return Qt.AlignmentFlag.AlignCenter | Qt.AlignmentFlag.AlignVCenter


class YFpCurrency(YFieldPainter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.attitude_color = get_attr(self, 'attitude_color')
        self.units = get_attr(self, 'units')
        self.zero = get_attr(self, 'zero', '0.00')
        if not self.units:
            self.units = 1

    def setup_painter(self, painter, value, text):
        color = self.get_color(value, text)
        if color:
            painter.setPen(color)

    def get_color(self, value, text):
        if not self.attitude_color:
            return None
        # 根据值设置颜色
        att = DisplayAttitude.INACTIVE
        if value > 0.001:
            att = DisplayAttitude.POSITIVE
        if value < -0.001:
            att = DisplayAttitude.NEGATIVE
        return self.get_attitude_color(att)

    def get_text(self, value, record):
        if value is None:
            return None
        v = to_float(value, none_value=0)
        if -0.0001 < v < 0.0001:  # 视为0
            return self.zero
        else:
            return f"{v / self.units:,.2f}"

    @staticmethod
    def alignment():
        return Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter


class YFpPercent(YFieldPainter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._color = current_theme.hover_color()
        self._max = 100.0
        self.zero = get_attr(self, 'zero', '0.0%')

    def draw_field(self, painter, rect, value, text):
        left_margin, right_margin = 4, 4
        full_width = rect.width() - left_margin - right_margin
        full_height = rect.height()
        top_margin, bottom_margin = full_height * 0.15, full_height * 0.15
        value = to_float(value)
        # 根据比率计算右边界
        right_margin = (self._max - value) / self._max * full_width + right_margin
        if right_margin > full_width:
            right_margin = full_width
        # 绘制百分比bar
        painter.setPen(self._color)
        brush = QBrush(self._color, Qt.BrushStyle.SolidPattern)
        painter.setBrush(brush)
        rect_bar = rect.adjusted(left_margin, top_margin, 0 - right_margin, 0 - bottom_margin)
        painter.drawRoundedRect(rect_bar, left_margin, top_margin)
        # 文字
        font = painter.font()
        font.setUnderline(True)  # 添加下划线
        painter.setFont(font)
        painter.setPen(Qt.GlobalColor.lightGray)
        painter.drawText(rect, Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignCenter, text)

    def get_text(self, value, record):
        if value is None:
            return None
        value = to_float(value)
        if -0.0001 < value < 0.0001:  # 视为0
            return self.zero
        else:
            return f"{value:,.1f}%"

    @staticmethod
    def alignment():
        return Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter


class YFpDateTime(YFieldPainter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_text(self, value, record):
        if pd.isna(value) or value is None:
            return None
        if isinstance(value, Timestamp):
            value = value.to_pydatetime()
        return f"{value}"

    @staticmethod
    def alignment():
        return Qt.AlignmentFlag.AlignCenter | Qt.AlignmentFlag.AlignVCenter


class YFpDate(YFieldPainter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_text(self, value, record):
        if pd.isna(value) or value is None:
            return None
        if isinstance(value, Timestamp):
            value = value.to_pydatetime()
        return format_date(value)

    @staticmethod
    def alignment():
        return Qt.AlignmentFlag.AlignCenter | Qt.AlignmentFlag.AlignVCenter


class YFpYearMonth(YFieldPainter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_text(self, value, record):
        return f'{value[:4]}年{value[4:]}月' if value else None

    @staticmethod
    def alignment():
        return Qt.AlignmentFlag.AlignCenter | Qt.AlignmentFlag.AlignVCenter


class YFpChoose(YFieldPainter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.code_dict = DictFetcher.instance().get_dict(self)
        self.attitude = get_attr(self, 'attitude')
        self.format = get_attr(self, 'format', '[{code}] - {name}')

    def get_color(self, value, text):
        if not self.attitude or pd.isna(value):
            return None
        for att, v in self.attitude.items():
            if value in v:
                return self.get_attitude_color(att)
        return None

    def setup_painter(self, painter, value, text):
        color = self.get_color(value, text)
        if color:
            painter.setPen(color)

    def get_text(self, value, record):
        ret = []
        values = value if is_collection(value) else [value]
        if self.code_dict:
            for code, name in self.code_dict.items():
                if str(code) in [str(v) for v in values]:  # 值和字典码均转换为字符串再做比较，防止值为数值但码表中的类型为字符时无法匹配
                    ret.append(eval(f'f"{self.format}"'))
        else:
            ret.extend([str(v) for v in values])
        return ', '.join(ret) if ret else None


class YFpLookup(YFieldPainter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.lookup_key = get_attr(self, 'lookup_key')
        self.lookup = FACTORY.get_lookup(**kwargs) if self.lookup_key else None

    def get_text(self, value, record):
        return self.lookup.to_text(value) if self.lookup else f"{value}"


#############################################################
# 数据字段准则控件
#############################################################
class YFdCriteria(QWidget):
    """
    数据字段的查询条件显示组件
    """

    def __init__(self,
                 column_spec,
                 parent=None,
                 editor_cls=None,
                 op: Op = None,
                 value=None,
                 layout_flag='H',
                 check_able=True,
                 auto_from=None,
                 ):
        super().__init__(parent=parent)
        self.column_spec = column_spec
        self.editor_cls = editor_cls if editor_cls else FACTORY.get_editor_class(self.column_spec)
        self.field = column_spec.name
        self.label = self.column_spec.title if self.column_spec.title else self.column_spec.comment
        if self.label is None:
            self.label = self.field.upper()
        self.check_able = check_able
        if self.label.startswith('*'):
            self.label = self.label[1:]
        self.layout_flag = layout_flag
        self.op = op if op else self.get_default_op()
        self._editors = []
        self._checked = False
        self.setup_ui()
        if auto_from:
            self.assign_from(auto_from)
        if value:
            self.set_value(value)
        self.layout().setContentsMargins(0, 0, 0, 0)

    def get_default_op(self):
        return {
            YFdText: Op.LIKE_IN,
            YFdNumber: Op.BETWEEN,
            YFdInt: Op.BETWEEN,
            YFdYear: Op.BETWEEN,
            YFdCurrency: Op.BETWEEN,
            YFdDate: Op.BETWEEN,
            YFdDateTime: Op.BETWEEN,
            YFdYearMonth: Op.BETWEEN,
            YFdChoose: Op.IN,
            YFdLookup: Op.IN,
        }[self.editor_cls]

    def add_to(self, layout: QLayout, index=-1):
        """
        将控件加入到指定的布局中
        :param index:
        :param layout: 待加入的布局对象
        :return:
        """
        label_text = f"{self.label} ({self.op.text()}):" if self.label else ""
        if isinstance(layout, QFormLayout):  # Form布局，label为checkbox
            w = QCheckBox(label_text, parent=self.parent())
            w.setChecked(self._checked)
            w.clicked.connect(self.on_check)
            layout.addRow(w, self)
        else:
            w = self
            if label_text:
                if self.check_able:  # 创建组合框
                    w = QGroupBox(self.parent())
                    w.setTitle(label_text)
                    w.setLayout(QVBoxLayout())
                    w.setCheckable(True)
                    w.setChecked(self._checked)
                    w.clicked.connect(self.on_check)
                    w.layout().setContentsMargins(0, 0, 0, 0)
                    w.layout().addWidget(self)
                else:  # 水平标题+组件
                    w = create_layout_widget(parent=self.parent(), layout_flag='H')
                    w.layout().addWidget(QLabel(label_text))
                    w.layout().addWidget(self)
            add_to_layout(layout, w, index)

    def setup_ui(self):
        def create_editor():
            editor = FACTORY.get_editor(self, self.column_spec,
                                        Stage.IN_QUERY,
                                        op=self.op,
                                        **self.column_spec.__dict__)
            self.layout().addWidget(editor)
            return editor

        def clear_editors(layout):
            clear_layout(layout)
            self._editors.clear()

        if self.layout():
            clear_editors(self.layout())
        self.setLayout(create_layout(self.layout_flag))
        self._editors.append(create_editor())
        if self.is_editor_pair():
            self.layout().addWidget(QLabel('~', alignment=Qt.AlignmentFlag.AlignCenter | Qt.AlignmentFlag.AlignVCenter))
            self._editors.append(create_editor())

    def get_value(self):
        values = [ed.get_value() for ed in self._editors]
        value = values[0] if len(values) < 2 else values
        return value

    def assign_to(self, qc):
        if self._checked and qc:
            set_attr(qc, self.field, (self.op.value, self.get_value()))
        else:
            set_attr(qc, self.field, None)

    def assign_from(self, obj):
        if obj:
            o_value = get_attr(obj, self.field)
            if o_value is None:
                return
            if not is_collection(o_value):
                raise Exception('查询条件值不符合要求，应为集合类型，且对应两个元素： (Op,Value)')
            new_op = Op(o_value[0])
            if new_op != self.op:  # 重置界面
                self.op = new_op
                self.setup_ui()
            self.set_value(o_value[1])

    def is_checked(self):
        return self._checked

    def on_check(self):
        sender = self.sender()
        self._checked = sender.isChecked()

    def is_editor_pair(self):
        return self.op == Op.BETWEEN

    def set_value(self, value):
        values = value if self.is_editor_pair() else [value]
        for i, edt in enumerate(self._editors):
            edt.set_value(values[i])
            if values[i] is not None:  # 如果存在有效值，自动设置条件生效标记
                self._checked = True


#####################################################################
# 字段相关组件工厂
#####################################################################
class YFdFactory:
    # 保存Lookup工具的静态字典
    _LOOKUPS = {}

    editor_map = {
        UnitType.TEXT: YFdText,
        UnitType.INT: YFdInt,
        UnitType.SWITCH: YFdSwitch,
        UnitType.NUMBER: YFdNumber,
        UnitType.CURRENCY: YFdCurrency,
        UnitType.DATE: YFdDate,
        UnitType.YEAR: YFdYear,
        UnitType.YEAR_MONTH: YFdYearMonth,
        UnitType.TIME: YFdDateTime,
        UnitType.CHOOSE: YFdChoose,
        UnitType.LOOKUP: YFdLookup,
        UnitType.PERCENT: YFdNumber,
        UnitType.RICH_TEXT: YFdRichText,
    }

    painter_map = {
        UnitType.TEXT: YFpText,
        UnitType.INT: YFpInt,
        UnitType.SWITCH: YFpSwitch,
        UnitType.NUMBER: YFpText,
        UnitType.CURRENCY: YFpCurrency,
        UnitType.DATE: YFpDate,
        UnitType.TIME: YFpDateTime,
        UnitType.YEAR: YFpText,
        UnitType.YEAR_MONTH: YFpYearMonth,
        UnitType.CHOOSE: YFpChoose,
        UnitType.LOOKUP: YFpLookup,
        UnitType.PERCENT: YFpPercent,
        UnitType.RICH_TEXT: YFpText,
    }

    @staticmethod
    def deduce_unit_type(col_spec) -> UnitType:
        # 简单使用col_spec确定的utype
        utype = get_attr(col_spec, 'utype')
        if utype:  # 指定了明确的utype，直接返回
            set_attr(col_spec, 'utype', utype)
            return utype
        # 没有明确指定utype,根据类型推断
        utype = UnitType.TEXT  # 默认TEXT
        dtype = get_attr(col_spec, 'dtype')
        if dtype:
            for p, u_type in {
                r'(NUMERIC)|(DECIMAL)': UnitType.CURRENCY,
                r'(INT.*)': UnitType.INT,
                r'(DATETIME.*)': UnitType.DATE,  # DATETIME类型默认还是用DATE
                r'(DATE *)': UnitType.DATE,
                r'(.*LOB.*)|(TEXT)': UnitType.RICH_TEXT,
            }.items():
                if re.match(p, dtype.upper()):
                    utype = u_type
        set_attr(col_spec, 'utype', utype)
        return utype

    @staticmethod
    def unite_dict(*ods) -> dict:
        d = dict()
        for o in ods:
            copy_obj(d, o)
        return d

    def get_editor_class(self, col_spec: ColumnSpec):
        utype = self.deduce_unit_type(col_spec)
        return self.editor_map[utype]

    def get_editor(self, parent, col_spec: ColumnSpec, stage: Stage = Stage.IN_TABLE, **kwargs):
        cls = self.get_editor_class(col_spec=col_spec)
        if cls is None:
            raise Exception(f'无法创建FieldEditor-找不到对应的 YFieldEditor Class,col_spec={col_spec}')
        return cls(parent=parent, stage=stage, **self.unite_dict(col_spec, kwargs))

    def get_painter(self, col_spec: ColumnSpec, **kwargs):
        utype = self.deduce_unit_type(col_spec)
        cls = self.painter_map[utype]
        if cls is None:
            raise Exception(f'无法创建FieldPainter-找不到对应的 YFieldPainter Class,col_spec={col_spec}')
        return cls(**self.unite_dict(col_spec, kwargs))

    @staticmethod
    def get_lookup(**kwargs):
        def split_r(s):
            dot_index = s.rfind('.')
            if dot_index != -1:
                return s[:dot_index], s[dot_index + 1:]
            else:  # 如果没有找到'.'，则返回整个字符串作为第一部分，第二部分为空字符串
                return s, ""

        lookup_key = get_attr(kwargs, 'lookup_key')
        if not lookup_key:
            raise Exception('实例化YLookup失败', '缺少lookup_key属性')
        lk = YFdFactory._LOOKUPS.get(lookup_key)
        if lk is None:
            mod, cls = split_r(lookup_key)
            # print(f'[YField] - YFdFactory.get_lookup: Creating instance of YLookup ... mod={mod},cls={cls}')
            if not mod or not cls:
                raise Exception("实例化YLookup失败",
                                f"无效的lookup key: '{lookup_key}' , 正确的格式应该为'.'连接的模块名和YLookup的子类名。")
            lk = create_instance(mod, cls, **kwargs)
            YFdFactory._LOOKUPS[lookup_key] = lk
        return lk


# 字段工具工厂
FACTORY = YFdFactory()



# -*- __rmi__.py: python ; coding: utf-8 -*-
# 访问服务器相关的工具类
import datetime
import time
from urllib.parse import urlparse

import requests
import urllib3

from . import Obj, d2o
from . import json_dump
from .win32 import get_profile, read_settings


def time_stamp(t) -> int:
    return int(time.mktime(t.timetuple()))


"""
http 请求-解析相关的工具类
"""


class RCall:
    service_base = None
    _default_obj_cls = Obj

    def __init__(self, url, direct_response=False):
        """
        远程调用
        :param url: 服务位置URL
        :param direct_response: 是否直接返回response对象。
        - True 直接返回response, get / post / put 方法中的cls参数将被忽略;
        - False 返回转换后的结果，类型由get / post / put 方法中的cls参数决定.
        """
        self._set_url(url)
        self._verify_ssl = False
        self._direct_response = direct_response  # 直接返回响应对象，不做json转换
        self._cert = ('', '')
        self._headers = self.default_headers()
        urllib3.disable_warnings()

    def default_headers(self):
        return {}

    def _set_url(self, url):
        if self.service_base:
            self._url = self.service_base + url
        else:
            self._url = url

    def _json_obj(self, r, obj_cls=None):
        if r.status_code != 200:
            raise RuntimeError(r.status_code, r.text)
        try:
            o = r.json(strict=False)
            o_type = type(o)
            if obj_cls is None:
                obj_cls = self._default_obj_cls
            if o_type in (list, tuple, set):
                return [d2o(oo, obj_cls) for oo in o]
            else:
                return d2o(o, obj_cls)
        except Exception as e:
            print('解析json错误', e)
            print(r.content)
            raise RuntimeError(e)

    def get(self, obj_cls=None, **kwargs):
        r = requests.get(self._url, params=kwargs, headers=self._headers, verify=self._verify_ssl, cert=self._cert)
        try:
            if self._direct_response:
                return r
            else:
                return self._json_obj(r, obj_cls)
        except Exception as e:
            print('error occurs while parsing json from request:', 'url=%s' % self._url, r, obj_cls, sep='\n')
            raise e

    def post(self, data=None, obj_cls=None, files=None, json=None, **kwargs):
        if data or json:
            json_data = json if json else json_dump(data)  # 给了json优先
            r = requests.post(self._url, json=json_data, headers=self._headers, files=files, params=kwargs,
                              verify=self._verify_ssl,
                              cert=self._cert)
        else:
            r = requests.post(self._url, files=files, headers=self._headers, params=kwargs, verify=self._verify_ssl,
                              cert=self._cert)
        if self._direct_response:
            return r
        else:
            return self._json_obj(r, obj_cls)

    def put(self, data=None, obj_cls=None, json=None, **kwargs):
        json_data = json if json else json_dump(data)  # 给了json优先
        r = requests.put(self._url, json=json_data, headers=self._headers, params=kwargs, verify=self._verify_ssl,
                         cert=self._cert)
        if self._direct_response:
            return r
        else:
            return self._json_obj(r, obj_cls)

    def delete(self, data=None, json=None, **kwargs):
        json_data = json if json else json_dump(data)  # 给了json优先
        r = requests.delete(self._url, json=json_data, headers=self._headers, params=kwargs,
                            verify=self._verify_ssl,
                            cert=self._cert)
        if self._direct_response:
            return r
        else:
            return self._json_obj(r)

    def url(self):
        return self._url


# 远程受保护的API调用
# 按照远程服务器的安全认证机制处理header，如果还没有获得服务器端的令牌，调用登录机制获得。
class AuthCall(RCall):
    EXPIRE_MINUTES = 20
    _secret_header = 'Token'
    _auth_token = None
    _last_login_time = None

    def __init__(self, url, header_builder=None):
        super().__init__(url)
        self._headers = self.default_headers()
        if header_builder:
            header_builder(self._headers)

    def has_expired(self):
        if self._last_login_time is None:
            return True
        return time_stamp(datetime.datetime.now()) - self._last_login_time > AuthCall.EXPIRE_MINUTES * 60

    # 处理安全机制
    def _auth(self):
        try:
            if AuthCall._auth_token and not self.has_expired():  # 已经有token了且未过期
                return
            else:  # 执行login
                AuthCall._auth_token = self._login()
                self._headers = self.default_headers()
        except Exception as e:
            raise RuntimeError('安全检查异常', e)

    def _login(self):
        user, mail, reg_code = get_profile()
        auth_code = read_settings('reg.auth', '')
        # login 需要post  multipart / from-data类型数据
        # 有两种方式：
        # 手动组建form-data并修改headers
        # 通过files参数传递form-data，推荐此种方式
        # form.username - 用户名/注册邮箱
        # form.password - 用户口令/授权码
        # form.client_id - 客户端标识/注册设备号
        files = {
            'username': (None, mail),
            "password": (None, auth_code),
            "client_id": (None, reg_code),
            "client_secret": (None, user),
        }
        response = requests.post(self.login_url(), files=files, verify=False)
        if response.status_code == 200:
            obj = response.json()
            AuthCall._last_login_time = time_stamp(datetime.datetime.now())
            return obj['access_token']
        else:
            raise RuntimeError(response.status_code, response.text)

    def login_url(self):
        url = urlparse(self.url())
        return fr"{url.scheme}://{url.hostname}:{url.port}/login"

    def get(self, obj_cls=None, **kwargs):
        self._auth()
        return super().get(obj_cls, **kwargs)

    def post(self, data=None, obj_cls=None, files=None, **kwargs):
        self._auth()
        return super().post(data, obj_cls, files, **kwargs)

    def put(self, data=None, obj_cls=None, json=None, **kwargs):
        self._auth()
        return super().put(data, obj_cls, **kwargs)

    def delete(self, data=None, **kwargs):
        self._auth()
        return super().delete(data, **kwargs)

    def default_headers(self):
        if AuthCall._auth_token:
            token = AuthCall._auth_token
            return {'authorization': f'Bearer {token}'}
        else:
            return {}



# -*- qrc_tools.py: python ; coding: utf-8 -*-
import os
import re


def search_files(fn_pattern, folder, sort=True, topdown=False):
    ret = []
    walk_path = folder if folder is not None else '.'
    for (this_dir, subs, files) in os.walk(walk_path, topdown=topdown):
        for fn in files:
            if re.match(fn_pattern, fn):
                ret.append((this_dir, fn))
    return sorted(ret) if sort else ret


def build_images_qrc(qrc_file_name, res_files):
    file_names = {}
    sp = '  '
    nl = '\n'
    with open(qrc_file_name, 'w') as file:
        file.write('<RCC>\n')
        for rc_prefix, files in res_files.items():
            f_names = []
            file.write(f'{sp}<qresource  prefix = "{rc_prefix}">{nl}')
            for f_dir, f_name in search_files(files[1], files[0], topdown=False):
                file.write(fr'{sp}{sp}<file alias="{f_name}">{f_dir}\{f_name}</file>{nl}')
                f_names.append(f_name)
            file.write(f'{sp}</qresource>{nl}')
            file_names[rc_prefix] = f_names
        file.write(f'</RCC>{nl}')
    return file_names


def build_enum(py_file, rc_names: dict):
    texts = []
    with open(py_file, 'w') as file:
        file.write(f'# -*- py_file.py: python ; coding: utf-8 -*-' + '\n')
        # file.write('from enum import Enum\n\n\n')
        # file.write('class PnKey(Enum):\n')
        for rc_prefix, f_names in rc_names.items():
            var_prefix = rc_prefix[1:2].upper()
            for name in f_names:
                v = f"{var_prefix}_{name.split('.')[0].upper()}"
                file.write(fr'{v} = ":{rc_prefix}/{name}"' + '\n')
                texts.append(name)
        file.write('\n')
    return texts


if __name__ == '__main__':
    qrc = build_images_qrc('images.qrc', {'/pic': ('images\\png', r'.*\.png'),
                                          '/icon': ('images\\ico', r'.*\.ico'), })
    print(qrc)
    build_enum('img.py', rc_names=qrc)
